/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

//
// Created by webpigeon on 19/11/22.
//

#ifndef FGGL_ASSETS_PACKED_ADAPTER_HPP
#define FGGL_ASSETS_PACKED_ADAPTER_HPP

#include <stack>
#include <ranges>
#include <algorithm>

#include "fggl/assets/packed/direct.hpp"
#include "fggl/data/storage.hpp"

namespace fggl::assets {

	using ResourceType = util::OpaqueName<uint64_t, struct ResourceTypeStruct>;
	constexpr ResourceType from_mime(const char* mime) {
		return ResourceType::make( util::hash_fnv1a_64(mime) );
	};

	struct ResourceRecord {
		std::filesystem::path m_path;
		AssetID assetID;
		ResourceType m_fileType;
		AssetTypeID m_assetType;
	};

	struct ManifestHeader {
		uint64_t assetID;
		uint64_t fileType;
		uint64_t assetType;
		uint64_t stringSize;
	};

	[[maybe_unused]]
	inline bool NEEDS_CHECKIN(std::filesystem::path, MemoryBlock) {
		debug::error("attempted to load asset which does not have a valid checkin yet");
		return false;
	}

	/**
	 * Adapter for Raw Checkin.
	 *
	 * For debugging/development it's a pain to have to pack assets directly. Although its much slower, it can be useful
	 * to be able to load non-optimised formats at runtime. This adapter allows injecting these non-optimised formats
	 * into the production checkin system.
	 */
	class CheckinAdapted {
		public:
			constexpr const static modules::ModuleService service = modules::make_service("fggl::assets::checkin::debug");
			using FilePredicate = std::function<AssetTypeID(std::filesystem::path)>;
			using FileLoader = std::function<bool(std::filesystem::path, MemoryBlock& block)>;

			CheckinAdapted(data::Storage* storage, RawCheckin* checkSvc) : m_storage(storage), m_checkin(checkSvc) {};

			// asset loading
			void load(AssetID asset) {
				auto& assetRef = m_files.at(asset);
				auto& loader = m_loaders.at(assetRef.m_fileType);

				MemoryBlock block;
				auto result = loader(assetRef.m_path, block);
				if ( !result ) {
					return;
				}

				m_checkin->check(assetRef.assetID, assetRef.m_assetType, block);
			}

			inline bool exists(AssetID asset) const {
				return m_files.find(asset) != m_files.end();
			}

			inline std::filesystem::path getPath(AssetID asset) const {
				auto& file = m_files.at(asset);
				return file.m_path;
			}

			void discover( const char* packName, bool useCache = false, bool updateCache = true) {
				std::string packRoot = "packs/";
				auto packDir = m_storage->resolvePath( data::StorageType::Data, packRoot + packName );
				discover(packDir, useCache, updateCache);
			}

			// asset discovery
			void discover( std::filesystem::path& packDir, bool useCache=true, bool updateCache=false ) {
				if ( useCache && has_manifest(packDir)) {
					// check if we've cached the search
					load_manifest(packDir.filename(), packDir);
					return;
				}

				std::vector<AssetID> packFiles;
				std::stack< std::filesystem::path > paths;
				paths.push(packDir);

				while ( !paths.empty() ) {
					auto path = paths.top();
					paths.pop();

					if ( std::filesystem::is_directory(path) ) {
						std::ranges::for_each( std::filesystem::directory_iterator{path}, [&paths](auto& path) {
							paths.push(path);
						});
					} else if ( std::filesystem::is_regular_file(path) ) {
						process_file( path, packDir, packFiles );
					}
				}

				// update the cache and remember what pack maps to what asset
				if ( updateCache ) {
					save_manifest(packDir.filename(), packDir, packFiles);
				}
				m_packs[ packDir.filename() ] = packFiles;
			}

			inline void setLoader(ResourceType type, FileLoader loader, FilePredicate predicate = nullptr) {
				assert( loader != nullptr );
				m_loaders[type] = loader;

				if ( predicate != nullptr ) {
					m_predicates[type] = predicate;
				}
			}

		private:
			data::Storage* m_storage;
			RawCheckin* m_checkin;
			std::map<AssetID, ResourceRecord> m_files;

			std::map<ResourceType, FilePredicate> m_predicates;
			std::map<ResourceType, FileLoader> m_loaders;
			std::map<std::string, std::vector<AssetID>> m_packs;

			void process_file(std::filesystem::path path, std::filesystem::path packDir, std::vector<AssetID> packFiles) {
				for( auto& [rType, pred] : m_predicates ) {
					// check the predicate, is this valid?
					auto aType = pred(path);
					if ( aType != INVALID_ASSET_TYPE ) {
						// it was, so we can finish processing
						auto packName = packDir.filename();
						auto relPath = std::filesystem::relative(path, packDir);
						ResourceRecord rr{
							.m_path = path,
							.assetID = make_asset_id(packName, relPath.generic_string()),
							.m_fileType = rType,
							.m_assetType = aType,
						};
						m_files[rr.assetID] = rr;
						packFiles.push_back( rr.assetID );
						debug::trace("discovered {} ({}) from pack '{}'", rr.assetID.get(), relPath.c_str(), packName.c_str() );
						break;
					}
				}
			}

			inline bool has_manifest(const std::string& packName) {
				auto packManifest = m_storage->resolvePath( data::StorageType::Cache, packName + "_manifest.bin" );
				return std::filesystem::exists(packManifest);
			}

			void load_manifest_entry(FILE* file, const std::filesystem::path& packRoot) {
				// read our entry ( id, ftype, atype, pathLen )
				ManifestHeader header{};
				std::fread(&header, sizeof(ManifestHeader), 1, file);

				// read the relative asset path
				char* relPath = new char[header.stringSize + 1];
				std::fread( relPath, sizeof(char), header.stringSize, file );
				relPath[ header.stringSize + 1 ] = '\0';

				// calculate and verify path
				std::filesystem::path fullPath = packRoot / relPath;
				delete[] relPath;

				if ( !std::filesystem::exists(fullPath) ) {
					debug::warning("pack manifest for {} contained invalid path {}", packRoot.filename().c_str(), fullPath.c_str());
					return;
				}

				// entry seems valid, load it
				ResourceRecord rr {
					.m_path = fullPath,
					.assetID = AssetID::make(header.assetID),
					.m_fileType = ResourceType::make(header.fileType),
					.m_assetType = AssetTypeID::make(header.assetType)
				};
				m_files[ rr.assetID ] = rr;
				m_packs[ packRoot.filename() ].push_back( rr.assetID );
				debug::trace("discovered {} ({}) from pack {}", rr.assetID.get(), fullPath.c_str(), packRoot.filename().c_str() );
			}

			void load_manifest(const std::string& packName, const std::filesystem::path& packRoot) {
				auto packManifest = m_storage->resolvePath( data::StorageType::Cache, packName + "_manifest.bin" );
				if ( !std::filesystem::exists(packManifest) ) {
					return;
				}

				// open the manifest file and start extracting entries
				FILE* file = std::fopen(packManifest.c_str(), "r");
				if ( file == nullptr ) {
					debug::warning("error opening manifest: {}", packManifest.c_str());
					return;
				}

				// read the number of entries
				uint64_t entries{0};
				std::fread(&entries, sizeof(uint64_t), 1, file);
				for ( uint64_t i = 0; i < entries; ++i) {
					load_manifest_entry(file, packRoot);
				}

				std::fclose( file );
			}

			void save_manifest(const std::string& packName, const std::filesystem::path& packRoot, const std::vector<AssetID>& assets) {
				auto packManifest = m_storage->resolvePath( data::StorageType::Cache, packName + "_manifest.bin", true);

				FILE* file = std::fopen(packManifest.c_str(), "w");
				if ( file == nullptr) {
					debug::warning("error saving manifest {}, missing dir?", packManifest.c_str());
					return;
				}

				const uint64_t entries{ assets.size() };
				std::fwrite( &entries, sizeof(uint64_t), 1, file);

				// write the entries
				for ( uint64_t i = 0; i < entries; ++i ) {
					auto& assetID = assets[i];

					auto& assetRecord = m_files.at(assetID);
					auto relPath = std::filesystem::relative(assetRecord.m_path, packRoot);
					auto relPathStr = relPath.generic_string();

					ManifestHeader mh {
						.assetID = assetRecord.assetID.get(),
						.fileType = assetRecord.m_fileType.get(),
						.assetType = assetRecord.m_assetType.get(),
						.stringSize = relPathStr.size()
					};
					std::fwrite( &mh, sizeof(ManifestHeader), 1, file );
					std::fwrite( relPathStr.c_str(), sizeof(char), relPathStr.size(), file );
				}

				std::fclose(file);
			}
	};

}

#endif //FGGL_ASSETS_PACKED_ADAPTER_HPP
